/* Copyright (c) 2008 Newcode Technology, Inc.  All rights reserved.  */ 

#include <mts.h>

class MTSPoolObject
{
public:
  MTSPool* pool;   // MTSPool pointer, create in ctor, destroy in dtor
  MTSPoolObject() {pool = MTSPoolCreate ("", sizeof "", MTSPoolNotThreaded);
                   // std::cout<<"Created MTSPool: "<< static_cast <void*> (pool) << std::endl;
                  }
  ~MTSPoolObject() { MTSPoolDestroy (pool);
                    // std::cout<<"Destroy MTSPool: "<< static_cast <void*> (pool) << std::endl;
                   }

  // the following two methods are called from stl
  void* allocate(size_t n) const {
    // std::cout << "MTSPoolAlloc " << n << std::endl;
    return MTSPoolAlloc (pool, n);  // return block from MTSPool
  }
  // add a free if your pools are going to be long lived
  // otherwise don't bother since the destructor will unmap the entire space
  void deallocate(void* p) const { 
    // std::cout << "deallocate zilch" << std::endl;
    // MTSPoolFree (pool, p);
  }

  // these are general MTSPool functions
  void Free(void* p) const { 
    MTSPoolFree (pool, p);
  }
  void FreeAll (void) const {
    MTSPoolFreeAll (pool);
  }
  char* Statistics (struct mts_stats *s) const {
    return MTSPoolStats (pool, s);
  }


private:
  MTSPoolObject& operator=(const MTSPoolObject&);
  MTSPoolObject(const MTSPoolObject&);
};

template <class T> class MTSPoolAllocator {
public:
  typedef T                 value_type;
  typedef value_type*       pointer;
  typedef const value_type* const_pointer;
  typedef value_type&       reference;
  typedef const value_type& const_reference;
  typedef std::size_t       size_type;
  typedef std::ptrdiff_t    difference_type;
  const MTSPoolObject& po_;
  
  template <class U>
  struct rebind { typedef MTSPoolAllocator<U> other; };

  MTSPoolAllocator(const MTSPoolObject& po) : po_(po) { /* std::cout<<"MTSPoolAllocator()"<<std::endl; */}
  MTSPoolAllocator(const MTSPoolAllocator& rhs) : po_(rhs.po_) {/* std::cout<<"MTSPoolAllocator(const MTSPoolAllocator&)"<<std::endl; */}
  template <class U>
  MTSPoolAllocator(const MTSPoolAllocator<U>& rhs) : po_(rhs.po_) {/* std::cout<<"MTSPoolAllocator(const MTSPoolAllocator<U>&)"<<std::endl; */}
  ~MTSPoolAllocator() {/*std::cout<<"~MTSPoolAllocator()"<<std::endl;*/}

  pointer address(reference x) const { return &x; }
  const_pointer address(const_reference x) const { 
    return x;
  }

  pointer allocate(size_type n, const_pointer = 0) {
    void* p = po_.allocate(n * sizeof(T));
    if (!p)
      throw std::bad_alloc();
    return static_cast<pointer>(p);
  }

  void deallocate(pointer p, size_type) { po_.deallocate(p);  }

  size_type max_size() const { 
    return static_cast<size_type>(-1) / sizeof(T);
  }

  void construct(pointer p, const value_type& x) { 
    new(p) value_type(x);
  }
  void destroy(pointer p) { p->~value_type(); }

  

private:
  MTSPoolAllocator& operator=(const MTSPoolAllocator&);

};

template<> class MTSPoolAllocator<void>
{
  typedef void        value_type;
  typedef void*       pointer;
  typedef const void* const_pointer;

  template <class U>
  struct rebind { typedef MTSPoolAllocator<U> other; }; };


template <class T>
inline bool operator==(const MTSPoolAllocator<T>&, 
                       const MTSPoolAllocator<T>&) {
  return true;
}

template <class T>
inline bool operator!=(const MTSPoolAllocator<T>&, 
                       const MTSPoolAllocator<T>&) {
  return false;
}

