#ifndef MTS_MTS_H
#define MTS_MTS_H 1
#include <unistd.h>

/* Copyright (c) 1997-2013 NewCode Technology, Inc.  All rights reserved. 

   If you have any questions please email support@newcodeinc.com or 
   call Ze'ev at +1 917 573 5786
*/

#ifdef __cplusplus
extern "C" {
#endif

/* 

MTS implements the standard memory allocator interfaces. This file  
provides additional interfaces to MTS to further maximize performance and
functionality.

It's divided into the following sections:

I.   Environment Variables
II.  Error Handling
III. Memory mapping & memory defaults
IV.  Pools
V.   Statistics
VI.  Obsolete


The initial version of MTS was developed in 1988 for a C++ CAD/CAM program 
running on a 4 MB Xenix/386 machine. At the time we found that while the 300k 
LOC program ran fine for a while, its performance was inevitably lost due
to excessive paging.

This experience resulted in a specialized heap subsystem which has proven to be 
highly durable.  MTS stands for Memory Tuning System.

MTS is built much like a micro-kernel with multiple allocators built on top
of basic services that bind them together. The current architecture provides
for an individual virtual heap per thread that handles allocation sizes up to 4k.
As pages become empty they are recycled among all these thread bound heaps.

For allocations above 4k and lesss than the block allocator default 512k, 
a fixed number of virtual heaps are created, by default this number is equal 
to the number of processors. At runtime, threads are mapped to a specific heap 
which is thread safe and can accommodate multiple concurrent allocations/deallocations. 
These heaps do not share memory. At the expense of diminished parallelism, their number 
can be minimized, so that space reuse is maximized. 

In our experience, if space management becomes is an issue, then the 
application is a good candidate for using MTS pools.

MTS has always been a very fast allocator, but more importantly it
scales for apps that use gigabytes of memory. Currently, it isn't unusual for
customers to manage 256GB within a single application.

I. Environment Variables

export MTS_HEAP_COUNT = N              
Set the number of fixed virtual heaps to N where N > 0. Doesn't affect the per
thread heaps. Default count is min (#cpus, 16)

export MTS_BADFREE_EXIT=EXIT           
Causes MTS to abort() on a bad free.

export MTS_LARGE_SIZE_WARNING = N      
Issues warnings when malloc sizes > N where N must be > 60k

export MTS_PRINT_COUNTS = 1
Print stats on exit

export MTS_PRINT_STATS = N
print stats every N seconds to stdout

MTS_PAGE_RECYCLE_GLOBAL = N
N pages stay local to heap. Default is 8.
0 means that all pages are recycled immediatly when they become empty.

MTS_BLOCK_SIZE = S
S must be >= 64k. Default is 512k. Allocations of S or greater are directly
memory mapped and unmapped when free() is called. Due to syscalls, it's slow 
but large allocations tend to persist and when freed are unmapped from app space
for optimum reuse.

MTS_SPIN_COUNT = N
Change the default spin lock count before a yield is called.

MTS_HUGE_PAGES = 1
For RHEL 6 and higher use huge pages by calling mmap() with MAP_HUGETLB set.
Make sure sufficient quantity of HUGE PAGES are available from the OS.

MTS_MAP_LOCKED = 1
Use MAP_LOCKED for MTS calls to mmap. Needs special privliges.

MTS_MAP_POPULATE = 1
Use MAP_POPULATE for MTS calls to mmap. (we still see page faults when using this option)

MTS_PREFAULT_MEMORY = 1
Cause MTS to touch each mmaped page

MTS_MEMORY_CALLBACK_DISABLE = 1
Cause MTS to ignore registered memory callback.

II.  Error Handling 

*/

typedef enum MTSstatus {MTSerror, MTSok} MTSstatus;

typedef void (*mts_error_callback) (int err_, const char* s, int s_length);
void mts_set_error_callback (mts_error_callback);

/* Warnings (type 0) are only processed in the presence of a defined callback 
   function. All other messages are sent to stderr and an error_callback if one
   is registered. Error callbacks are also a convenient place to set a break-
   point to debug a memory related error.  

   Values for err_: 
   0 - warning 
   1 - non fatal error
   2 - error writing output
   3 - system call error
   4 - pool related errors
   5 - pool size exceeded limit

  -1 - bad free, non fatal by default

   MTS will call _exit on the following errors

  -2 - fatal error
  -3 - fatal error writing output
  -4 - fatal system call error
*/


void mts_warn_large_size (unsigned long Size);

/* Warns of large allocations. Size must be greater than 60k otherwise its
   silently ignored. This was incorporated to help find sporadic large
   allocations that often precede crashes. 
*/


/*
III. Memory mapping

Using a callback mechanism, MTS lets applications mmap the main allocator and 
individual pools of allocations. 
*/

typedef struct mts_pool MTSPool;

/* MTSPool* is an opaque pointer to an internal pool structure */

typedef void* (*mts_user_memory_callback_v2) (unsigned long size, void* user_data);
typedef void (*mts_user_unmap_memory_callback_v2) (void* sddrize, size_t length);

/* Signature of the user defined callback that provides MTS with memory space.
   It accepts a size to mmap and a user_data pointer provided by the 'set'
   callback function. It returns a valid mmaped address of size bytes 
   or MMAP_FAILED (-1). MTS takes care of page alignment for mmapped memory.  
*/

// added a handle for an munmap user callback
MTSstatus 
mts_set_memory_callback_v3 (MTSPool* p, 
                            mts_user_memory_callback_v2, 
                            mts_user_unmap_memory_callback_v2, 
                            void* user_data);

MTSstatus 
mts_set_memory_callback_v2 (MTSPool* p, 
                            mts_user_memory_callback_v2, 
                            void* user_data);

/* When p is null ->  set the general allocator memory callback, otherwise
   set the memory callback for a specific pool. user_data allows the callback to
   service calls for memory from many pools to different mmaped files.  

We recommend that the callback make a special effort to not call malloc
recursively which could lead to an infinite loop. One approach would be
to use thread local memory to guard against this:

This is the callback code you can find in the memory mapping example:

//! Prevent infinite recursion with a thread-local variable.
__thread bool allocatingMemory;

// return a valid mmap address or MAP_FAILED
void* myMTSmemoryCallback (size_t size, void* memFile)
{
  if (memFile == NULL || allocatingMemory)
  { return mmap(NULL, size, PROT_READ | PROT_WRITE, MAP_PRIVATE | MAP_ANONYMOUS, -1, 0);
  }
  else
  { allocatingMemory = true;
    void* ret = growMemMap(size, (MemoryFile*)memFile);		
    allocatingMemory = false;
    return ret;
  }
}
*/

// To initialize the size of a heap, two calls:

size_t mts_set_small_heap_size (size_t size, size_t inc, unsigned long opt);
size_t mts_set_large_heap_size (size_t size, size_t inc, unsigned long opt);

// opt

const unsigned long MTSSmallHeapRecycleOff =    0x0800;		// effects small heaps only
const unsigned long MTSPreFault =               0x1000;		// touches all newly mmaped pages
const unsigned long MTSPreFaultPrev =           0x4000;		// touches all previously mmaped pages
// mmap options for internal MTS calls
const unsigned long MTSMapHuge          =       0x10000;	// MAP_HUGETLB (RH6 and higher)
const unsigned long MTSMapLocked        =       0x20000;	// MAP_LOCKED
const unsigned long MTSMapPopulate      =       0x40000;	// MAP_POPULATE


/*
Both set size calls return the size of the heap. 'size' is the current size to set.
'inc' is the increment size to grow the heap. The calls were created to help prefault
memory to speed up application startup.

if MTSSmallHeapRecycleOff is set for the small heap call, page recycling is turned off
for that heap. This is often desirable if there is prior knowledge that the heap will
continously reuse the same sort of allocations. For the pages to migrate to other
heaps to be later reclaimed is usually wasteful.

When setting the size of large heaps, a private heap for allocations greater than 4k
is created which will be bound to the thread it is created in. Tt creates it. This heap will 
persist beyond the life of the thread that creates it and become orphaned. Use this 
call only for threads that will exist for the life of the application where you know 
that lots of memory is needed.  Otherwise, let the thread use it's internally assigned 
and initialized fixed heap.

IV.  Pools

MTS as of 7.9.2 incorporates two types of pool interfaces. A standard pool
interface and a light weight (lw) pool interface. The original standard  pool
interface incorporates all the traditional allocation/deallocation MTS
strategies, whereas, the lw pool interface just focuses on achieving very
high speed allocation results and uses a mass memory free. This is useful, for instance,
for allocations associated with parsing message buffers that are constantly being allocated 
and discarded. Both pool interfaces suppot the general call to free(v) where v is a
pool allocation. So existing code will work seamlessly with these allocations.

A. Standard MTS Pool Interface

MTS early in its inception incorporated a pool interface. The main
utility of MTS pools is that they aggregate a collection of allocations 
which can be deleted together and (unlike the default heaps) unmapped from 
the application.  It's important to note  that individual MTS pool allocations 
can be freed using an ordinary call to free(), so disparate parts of the 
application need not be aware of their origin.
*/

// MTSPool Create Options

const size_t MTSPoolNotThreaded       = 0x0000;		// default
const size_t MTSPoolThreaded          = 0x0008;
const size_t MTSPoolInternal          = 0x0010;		// default
const size_t MTSPoolPrivate           = 0x0020;
const size_t MTSPoolShared            = 0x0040;
const size_t MTSPoolAppUnmap          = 0x0080;
const size_t MTSPoolMTSUnmap          = 0x0100;		// default

/* These options control MTSPool attributes of locking, space and unmapping. 
   Note that locking refers to internal MTS locking not read/write locks.

   MTSPoolNotThreaded   Pool is not thread safe. 
   MTSPoolThreaded      Pool is thread safe
   
   MTSPoolInternal      MTS automatically maps space as the pool grows. 
   MTSPoolPrivate       Application provides mmapped memory
   MTSPoolShared        Application provides mmapped shared memory 
   
   MTSPoolAppUnmap      Application is responsible for unmapping pool memory
   MTSPoolMTSUnmap      MTSPoolDestroy will unmap pool memory. 
 

   MTSPoolCreatV3 creates a new named pool. 'name' is null terminated. The opt
   parameter is a bitmapped combination of MTSPool Options.

   If the pool type is MTSPoolInternal addr and size are ignored. MTS will 
   dynamically size the pool based on allocation/deallocation requests and 
   unmmap the space when MTSPoolDestroy is called. 

   For MTSPoolPrivate and MTSPoolShared pools, initial mmapped space with a 
   minimum of 256KB should be provided. If the application anticipates that 
   the initial space will become insufficient, a memory callback must be 
   registered.

   For MTSPoolShared a process can call MTSPoolCreatV3 with an already
   established pool in shared memory which MTS will recognize. To establish
   a fresh shared space for an MTS pool in shared memory the first 64k need 
   to be set to 0.

   If a process tries to use an existing shared pool that consists of more than
   one mmaped segment, it is the process responsibility to mmap those segments
   into its process space at the exact addresses, but only one call to 
   MTSPoolCreatV3 with the initial segment address and size should be made.

   When multiple processes are mapped to the same shared pool, if the pool
   is declared MTSPoolThreaded, the pool will be 'process safe' as well
   as thread safe.  
*/

MTSPool   *MTSPoolCreateV3  (const char* name, size_t opt, void* addr, size_t size);
MTSstatus  MTSPoolDestroy   (MTSPool*);					// depending on create settings, unmaps the pool
MTSstatus  MTSPoolAddSpace  (MTSPool* pool, void* addr, size_t size);	// increase the size of a Private or Shared pool
void      *MTSPoolAlloc     (MTSPool*, size_t);                     	// malloc inside a pool using MTSPool* returned by CreatV3
void      *MTSPoolRealloc   (MTSPool*, void *, size_t);             	// realloc inside a pool
void       MTSPoolFree      (MTSPool*, void*); 				// also free (void*) works 
MTSstatus  MTSPoolFreeAll   (MTSPool*);        				// free all allocations 
MTSPool   *MTSPoolGet       (void*);           				// identify pool allocation 
char      *MTSPoolGetName   (MTSPool*);        				// return pool name 
MTSstatus  MTSPoolSetLS     (MTSPool*, void *);				// store a private pointer 
void      *MTSPoolGetLS     (MTSPool*);        				// get it back 


/*

B. Light Weight Pool Interface

The lw pool interface is designed for speed. Speed of pool initialization
and alloc execution. In this scheme a call to free() is a nop.
All memory in the lw pool is freed by a call to Release the pool. 
Sizing the pool when pool Create is called will gaurantee the best 
performance.

The only size stat that is kept is the total amount of memory requested from 
the OS for this lw pool. No current used or malloc/free counts are possible.

Currently no internal locking is supported. If locking is needed, it's the application
responsibility.

*/

const size_t MTSlwPoolRecycle	= 0x8000;				// recycle lw pools instead of destroying them.

void * MTSlwPoolCreate  (const char * name, int opt, size_t size);	// returns lw ptr or 0
void * MTSlwPoolAlloc   (void* lwp, size_t size);			// allocate size bytes in lw pool
size_t MTSlwPoolRelease (void* lwp);					// release the lw pool for reuse, return ref count
int    MTSlwPoolDestroy (void* lwp);					// destroy and unmap the space
size_t MTSlwPoolSize    (void* lwp);					// lw pool size
char * MTSlwPoolGetName (void * lwp);					// get the name stored during Create.
void   MTSlwPoolSetLS   (void* lwp, void* ptr);				// store a pointer locally in the lw pool
void * MTSlwPoolGetLS   (void* lwp);					// retrieve a local pointer
/*

V.  Statistics

   new for version 7.10.2 are thread based statistics. See below.

   MTS provides a comprehensive array of internal statistics.
   Collecting statistics imposes a ~5% performance penalty. 
   If you would like to deploy a version of MTS that doesn't collect 
   stats, please let us know.

   The small_arena count is the total number of thread specific heaps, which 
   includes both those that are currently associated with an active
   thread and those that are inactive, i.e., their thread has exited. 
   Inactive thread heaps may still contain allocations that are accessed
   from active threads.
 
   The size metrics are mmaped memory sizes whereas the byte count metric 
   refers to how much space is currently in use.

   Since the thread heaps recycle pages among themselves the D_free_pages
   tracks how much space is available and unused for the thread heaps.

*/

struct mts_stats {
   size_t heap_size;          /* total heap size as requested from OS */
   size_t byte_count;         /* current byte usage */
   size_t small_malloc;	      /* number of calls for small allocations */
   size_t medium_malloc;      /* number of calls for medium allocations */
   size_t large_malloc;	      /* number of calls for large allocations */
   size_t small_free;	      /* number of calls for small deallocations */
   size_t medium_free;	      /* number of calls for medium deallocations */
   size_t large_free;;	      /* number of calls for large deallocations */
   size_t unused6;            
   size_t pools;              /* number of pools */
   size_t small_arenas;       /* number of dynamic heaps for sizes <=4096 */
   size_t yield_count;        /* sched_yield count for malloc/free */
   size_t heap_count;         /* number of fixed heaps for sizes > 4096 */
   size_t F_size;             /* fixed heaps size */
   size_t F_byte_count;       /* fixed heaps currently in use*/
   size_t DA_size;            /* active dynamic thread heaps size */
   size_t DA_byte_count;      /* active dynamic thread heaps current bytes */
   size_t DI_size;            /* inactive dynamic thread heaps size */
   size_t DI_byte_count;      /* inactive dynamic thread heaps current bytes */
   size_t D_free_pages;       /* dynamic thread heaps free pages in bytes */
   size_t P_size;             /* Pools size */
   size_t P_byte_count;       /* Pool current bytes */
   size_t list_yield_count;   /* dynamic thread heap inactive list lock */
   size_t page_yield_count;   /* dynamic thread heap free page lock */
   };

int mts_total_heap_stats1 (struct mts_stats *);     /* returns 0 on success */
char *MTSPoolStats (MTSPool* p, struct mts_stats *s);  /* returns pool name */

struct mts_pthread_stats {
        pid_t    tid;		/* tid of pthread allocating the space */
        size_t  space;		/* total space requested by this thread */
        size_t  small_used;	/* current small allocation */
        size_t  middle_used;	/* current middle size allocations */
        size_t  large_used;	/* current large allocations */
        size_t  block_used;	/* current mmaped block allocations */
};				/* sum of used is current allocations made by this pthread */

struct mts_pthread_stats* mts_collect_pthread_stats (void);		// all past and present pthread stats
struct mts_pthread_stats* mts_collect_active_pthread_stats (void);
struct mts_pthread_stats* mts_collect_inactive_pthread_stats (void);	

int mts_check_consistency(int options);		// no options yet

const char * mts_version(void);			// return mts version number
/*

VI.   Obsolete

   These functions are still maintained but we recommend using replacement
   functions from the previous sections.
*/


typedef void* (*mts_user_memory_callback) (size_t length);
void mts_set_memory_callback (mts_user_memory_callback);

int mts_total_heap_stats (struct mts_stats *);          

unsigned long mts_heapbytecounts(void);
unsigned long mts_total_heapsize(void);

/* returns the number of virtual heaps.  */
int mts_num_heaps ();

/* sets the number of heaps, overridden by MTS_HEAP_COUNT and MTS_INIT_THREAD_HEAPS env variables */
void mts_init_thread_heaps( int np );

/* return the index of the heap that the current thread is using */
int mts_which_heap();


/* heapsize functions return the size of the virtual heap(s):
 *
 * mts_heapsizes:               sum of all heap memory gotten from the OS
 * mts_current_heapsize:        size of virtual heap this thread is using
 * mts_heapsize:                size of a particular heap indexed
 *                              0 to (mts_num_heaps() -1)
*/
unsigned long mts_heapsizes();
unsigned long mts_current_heapsize ();
unsigned long mts_heapsize ( int heap );

/* heapbytecount functions return current malloc'ed memory.
 * NOTE: more than one thread can be bound to a heap, therefor these stats
 * can reflect the allocations of more than one single thread.
 *
 * mts_heapbytecounts:          sum of all currently malloced memory
 * mts_current_heapbytecount:   malloc'ed memory in heap that thread is using
 * mts_heapbytecount:           malloc'ed memory of a particular heap indexed
 *                              0 to (mts_num_heaps() -1)
*/
unsigned long mts_heapbytecounts ();
unsigned long mts_current_heapbytecount();
unsigned long mts_heapbytecount ( int heap );



#ifdef __cplusplus
}
#endif
#endif

